import datetime
from chardet import detect as chdetect

import proxysiphon.proxychimp as proxychimp


def read_ncdc(filepath_or_buffer, encoding=None):
    """Read NOAA NCDC txt file"""
    with open(filepath_or_buffer, 'rb') as fl:
        flbytes = fl.read()
    if encoding is None:
        encoding = chdetect(flbytes)['encoding']
    filestr = flbytes.decode(encoding)
    return NcdcRecord(filestr)


def find_values(x, k, sep=':', fun=None):
    """Find values in a list of strings containing [k][sep][values]

    Parameters
    ----------
    x : iterable
        Iterable of strings representing lines in a file.
    k : str
        A key or pattern that to search x for.
    sep : str
        A separator that divides the `k` pattern from the target value.
    fun : function-like, optional
        Function used to format target value before returning.

    Returns
    -------
    The target value is returned after removing excess whitespace from left
    and right of the value. If the key or target pattern is not found,
    then None is returned. If an empty string, or whitespace is the found
    value, then None is returned.
    """
    out = None
    for l in x:
        if k in l and sep in l:
            val = l.split(sep, maxsplit=1)[1:][0].lstrip().rstrip()
            if val != '':
                out = val
    if fun is not None and out is not None:
        out = fun(out)
    return out


class SiteInformation:
    def __init__(self, s):
        self.site_name = None
        self.location = None
        self.country = None
        self.northernmost_latitude = None
        self.southernmost_latitude = None
        self.easternmost_longitude = None
        self.westernmost_longitude = None
        self.elevation = None

        s = str(s)
        s = s.splitlines()
        self.site_name = find_values(s, '# Site_Name:', fun=str)
        self.location = find_values(s, '# Location:', fun=str)
        self.country = find_values(s, '# Country:', fun=str)
        self.northernmost_latitude = find_values(s, '# Northernmost_Latitude:', fun=float)
        self.southernmost_latitude = find_values(s, '# Southernmost_Latitude:', fun=float)
        self.easternmost_longitude = find_values(s, '# Easternmost_Longitude:', fun=float)
        self.westernmost_longitude = find_values(s, '# Westernmost_Longitude:', fun=float)
        self.elevation = find_values(s, '# Elevation:', fun=float)


class DataCollection:
    def __init__(self, s):
        self.collection_name = None
        self.first_year = None
        self.last_year = None
        self.time_unit = None
        self.core_length = None
        self.notes = None
        self.collection_year = None

        s = str(s)
        s = s.splitlines()
        self.collection_name = find_values(s, 'Collection_Name:', fun=str)
        self.first_year = find_values(s, 'First_Year:', fun=float)
        self.last_year = find_values(s, 'Last_Year:', fun=float)
        self.time_unit = find_values(s, 'Time_Unit:', fun=str)
        self.core_length = find_values(s, 'Core_Length:', fun=str)
        self.notes = find_values(s, 'Notes:', fun=str)
        self.collection_year = find_values(s, 'Collection_Year:', fun=int)


class ChronologyInformation:
    def __init__(self, s):
        self.df = None
        if len(s) == 2:
            s_describe = s[0]
            self.df = s[1].copy()


class Data:
    def __init__(self, s):
        self.df = None
        if len(s) == 2:
            s_describe = s[0]
            self.df = s[1].copy()


class Publication:
    def __init__(self, s):
        self.doi = None

        s = str(s)
        s = s.splitlines()
        self.doi = find_values(s, '# DOI:', fun=str)


class ContributionDate:
    def __init__(self, s):
        self.date = None

        s = str(s)
        s = s.splitlines()
        try:
            v = find_values(s, 'Date:', fun=str).split('-')
            assert len(v) == 3
            self.date = datetime.date(year=int(v[0]), month=int(v[1]), day=int(v[2]))
        except AttributeError:
            # Date information not found, as v is None, as is self.date
            pass


class NcdcRecord:
    def __init__(self, s):
        g = proxychimp.Guts(s)

        def stringer(guts, sec):
            try:
                out = '\n'.join(guts.pull_section(sec))
            except KeyError:
                out = None
            return out

        self.contribution_date = ContributionDate(stringer(g, 'Contribution_Date'))
        self.publication = Publication(stringer(g, 'Publication'))
        self.site_information = SiteInformation(stringer(g, 'Site Information'))
        self.chronology_information = ChronologyInformation([stringer(g, 'Chronology_Information'), g.yank_chron_df()])
        self.data = Data([stringer(g, 'Data'), g.yank_data_df()])
        self.data_collection = DataCollection(stringer(g, 'Data_Collection'))



from tempfile import NamedTemporaryFile
import pytest
import pandas as pd

from proxysiphon import proxychimp


datapayload = ['#------------------------',
           '# Contribution_Date',
           '#',
           '#------------------------',
           '# Title',
           '#',
           '#------------------------',
           '# Chronology_Information',
           '#',
           '# Labcode\tdepth_top\tdepth_bottom\tmat_dated\t14C_date\t14C_1s_err\tdelta_R\tdelta_R_1s_err\tother_date\tother_1s_err\tother_type\t',
           '# 152757	5	6	G. ruber or mixed planktonic	475	30	188	73	-999	-999	-999	',
           '#',
           '#------------------------',
           '# Data',
           '# Data lines follow (have no #)',
           '# Missing Value: -999',
           'depth\tage\tbacon',
           '1\t2\t3',
           '4\t5\t6']


@pytest.fixture(scope='module')
def chron_nodeltaR_nodata_guts():
    payload = ['#------------------------',
               '# Contribution_Date',
               '#',
               '#------------------------',
               '# Title',
               '#',
               '#------------------------',
               '# Chronology_Information',
               '#',
               '# Labcode\tdepth_top\tdepth_bottom\tmat_dated\t14C_date\t14C_1s_err\tdelta_R\tdelta_R_1s_err\tother_date\tother_1s_err\tother_type\t',
               '# 152757	5	6	G. ruber or mixed planktonic	-999	-999	-999	-999	-999	-999	-999	',
               '#',
               '#------------------------',
               '# Data',
               '# Data lines follow (have no #)',
               '# Missing Value: -999',
               'depth\tage\tbacon']
    with NamedTemporaryFile('wb') as tf:
        tf.write('\n'.join(payload).encode('utf-8'))
        tf.flush()
        g = proxychimp.Guts(tf.name)
    return g


@pytest.fixture(scope='module')
def chron_guts():
    payload = datapayload
    with NamedTemporaryFile('wb') as tf:
        tf.write('\n'.join(payload).encode('utf-8'))
        tf.flush()
        g = proxychimp.Guts(tf.name)
    return g


@pytest.fixture(scope='module')
def dumb_guts():
    payload = ['#------------------------',
               '# Contribution_Date',
               '#',
               '#------------------------',
               '# Title',
               '#',
               '#------------------------',
               '# Chronology_Information',
               '#',
               '# Labcode\tdepth_top\tdepth_bottom\tmat_dated\t14C_date\t14C_1s_err\tdelta_R\tdelta_R_1s_err\tother_date\tother_1s_err\tother_type\t',
               '#',
               '#------------------------',
               '# Data',
               '# Data lines follow (have no #)',
               '# Missing Value: -999',
               'depth\tage\tbacon',
               '1\t2\t3',
               '4\t5\t6']
    with NamedTemporaryFile('wb') as tf:
        tf.write('\n'.join(payload).encode('utf-8'))
        tf.flush()
        g = proxychimp.Guts(tf.name)
    return g


def test_str_guts__init_():
    filestr = '\n'.join(datapayload)
    g = proxychimp.Guts(filestr)
    goal_dict = {'Labcode': [152757], 'depth_top': [5], 'depth_bottom': [6],
                 'mat_dated': ['G. ruber or mixed planktonic'],
                 '14C_date': [475],'14C_1s_err': [30],
                 'delta_R': [188], 'delta_R_1s_err': [73],
                 'other_date': [pd.np.nan], 'other_1s_err': [pd.np.nan],
                 'other_type': [pd.np.nan]}
    goal = pd.DataFrame(goal_dict)
    df = g.yank_chron_df()
    for k in goal_dict.keys():
        pd.testing.assert_series_equal(goal[k], df[k])


def test__divide_portions(dumb_guts):
    goal_data = ['depth\tage\tbacon', '1\t2\t3', '4\t5\t6']
    goal_description_ends = ('#------------------------', '# Missing Value: -999')
    goal_beginline = 15
    assert dumb_guts.data == goal_data
    assert dumb_guts.description[0] == goal_description_ends[0]
    assert dumb_guts.description[-1] == goal_description_ends[-1]
    assert dumb_guts.data_beginline == goal_beginline


def test__write_sectionindex(dumb_guts):
    goal_index = {'Chronology_Information': (7, 11),
                  'Contribution_Date': (1, 3),
                  'Data': (12, 15),
                  'Title': (4, 6)}
    assert len(dumb_guts.sectionindex.keys()) == 4
    for k, goal in goal_index.items():
        assert dumb_guts.sectionindex[k] == goal


def test_pull_section(dumb_guts):
    goal = ['# Data', '# Data lines follow (have no #)', '# Missing Value: -999']
    assert dumb_guts.pull_section('Data') == goal


def test_pull_section_badname(dumb_guts):
    with pytest.raises(KeyError):
        dumb_guts.pull_section('data')


def test_available_sections(dumb_guts):
    goal = ['Contribution_Date', 'Title', 'Chronology_Information', 'Data']
    assert dumb_guts.available_sections() == goal


def test_guess_missingvalues(dumb_guts):
    goal = '-999'
    assert dumb_guts.guess_missingvalues() == goal


def test_yank_data_df(dumb_guts):
    goal_dict = {'depth': [1, 4], 'age': [2, 5], 'bacon': [3, 6]}
    goal = pd.DataFrame(goal_dict)
    df = dumb_guts.yank_data_df()
    for k in goal_dict.keys():
        pd.testing.assert_series_equal(goal[k], df[k])


def test_yank_chron_df(chron_guts):
    goal_dict = {'Labcode': [152757], 'depth_top': [5], 'depth_bottom': [6],
                 'mat_dated': ['G. ruber or mixed planktonic'],
                 '14C_date': [475],'14C_1s_err': [30],
                 'delta_R': [188], 'delta_R_1s_err': [73],
                 'other_date': [pd.np.nan], 'other_1s_err': [pd.np.nan],
                 'other_type': [pd.np.nan]}
    goal = pd.DataFrame(goal_dict)
    df = chron_guts.yank_chron_df()
    for k in goal_dict.keys():
        pd.testing.assert_series_equal(goal[k], df[k])


def test_has_chron(dumb_guts, chron_guts):
    assert dumb_guts.has_chron() is False
    assert chron_guts.has_chron() is True


def test_has_deltar(dumb_guts, chron_guts, chron_nodeltaR_nodata_guts):
    assert chron_nodeltaR_nodata_guts.has_deltar() is False
    assert dumb_guts.has_deltar() is False
    assert chron_guts.has_deltar() is True


def test_has_datacolumn(dumb_guts, chron_nodeltaR_nodata_guts):
    assert chron_nodeltaR_nodata_guts.has_datacolumn('boogers') is False
    assert dumb_guts.has_datacolumn('depth') is True



import pytest
import pandas as pd

from proxysiphon import records, proxychimp


@pytest.fixture(scope='module')
def testchrondatalines():
    lines = ['#------------------------',
               '# Contribution_Date',
               '#',
               '#------------------------',
               '# Title',
               '#',
               '#------------------------',
               '# Chronology_Information',
               '#',
               '# Labcode\tdepth_top\tdepth_bottom\tmat_dated\t14C_date\t14C_1s_err\tdelta_R\tdelta_R_1s_err\tother_date\tother_1s_err\tother_type\t',
               '# 152757	5	6	G. ruber or mixed planktonic	475	30	188	73	-999	-999	-999	',
               '#',
               '#------------------------',
               '# Data',
               '# Data lines follow (have no #)',
               '# Missing Value: -999',
               'depth\tage\tbacon',
               '1\t2\t3',
               '4\t5\t6']
    return '\n'.join(lines)


def test_find_values():
    lines = ['apple: 1\n', 'bee: 1:2\n', 'charlie: 1.5\n', 'dingo-2\t\n', 'echo: ']
    assert records.find_values(lines, 'NotAKey') is None
    assert records.find_values(lines, 'apple') == '1'
    assert records.find_values(lines, 'bee') == '1:2'
    assert records.find_values(lines, 'charlie', fun=float) == 1.5
    assert records.find_values(lines, 'bee', sep='-', fun=int) is None
    assert records.find_values(lines, 'dingo', sep='-', fun=int) == 2
    assert records.find_values(lines, 'echo') is None


def test_SiteInformation():
    testlines = """# Site Information
                # Site_Name: P178-15P
                # Location: Arabian Sea
                # Country: 
                # Northernmost_Latitude: 11.955
                # Southernmost_Latitude: 11.955
                # Easternmost_Longitude: 44.3
                # Westernmost_Longitude: 44.3
                # Elevation: -869"""
    v = records.SiteInformation(testlines)
    assert v.site_name == 'P178-15P'
    assert v.location == 'Arabian Sea'
    assert v.country is None
    assert v.northernmost_latitude == 11.955
    assert v.southernmost_latitude == 11.955
    assert v.easternmost_longitude == 44.3
    assert v.westernmost_longitude == 44.3
    assert v.elevation == -869


def test_DataCollection():
    testlines = """# Data_Collection
                #   Collection_Name: P178-15P
                #   First_Year: 39485
                #   Last_Year: -18
                #   Time_Unit: cal yr BP
                #   Core_Length: 
                #   Notes: mg_red
                #   Collection_Year: 1923"""
    v = records.DataCollection(testlines)
    assert v.collection_name == 'P178-15P'
    assert v.first_year == 39485
    assert v.last_year == -18
    assert v.time_unit == 'cal yr BP'
    assert v.core_length is None
    assert v.notes == 'mg_red'
    assert v.collection_year == 1923


def test_ChronologyInformation(testchrondatalines):
    goal_dict = {'Labcode': [152757], 'depth_top': [5], 'depth_bottom': [6],
                 'mat_dated': ['G. ruber or mixed planktonic'],
                 '14C_date': [475],'14C_1s_err': [30],
                 'delta_R': [188], 'delta_R_1s_err': [73],
                 'other_date': [pd.np.nan], 'other_1s_err': [pd.np.nan],
                 'other_type': [pd.np.nan]}
    goal = pd.DataFrame(goal_dict)
    g = proxychimp.Guts(testchrondatalines)
    df = records.ChronologyInformation(['\n'.join(g.pull_section('Data')), g.yank_chron_df()]).df
    for k in goal_dict.keys():
        pd.testing.assert_series_equal(goal[k], df[k])


def test_Data(testchrondatalines):
    goal_dict = {'depth': [1, 4], 'age': [2, 5], 'bacon': [3, 6]}
    goal = pd.DataFrame(goal_dict)
    g = proxychimp.Guts(testchrondatalines)
    df = records.Data(['\n'.join(g.pull_section('Data')), g.yank_data_df()]).df
    for k in goal_dict.keys():
        pd.testing.assert_series_equal(goal[k], df[k])


def test_Publication():
    testlines = """# Publication
                # Authors: Tierney, Jessica E.; Pausata, Francesco S. R.; deMenocal, Peter B.
                # Published_Date_or_Year: 2016
                # Published_Title: Deglacial Indian monsoon failure and North Atlantic stadials linked by Indian Ocean surface cooling
                # Journal_Name: Nature Geoscience
                # Volume: 9
                # Edition: 
                # Issue: 
                # Pages: 46-50
                # Report Number: 
                # DOI: 10.1038/ngeo2603
                # Online_Resource: 
                # Full_Citation: 
                # Abstract: """
    v = records.Publication(testlines)
    assert v.doi == '10.1038/ngeo2603'


def test_ContributionDate():
    testlines = """# Contribution_Date
                #    Date: 2017-03-31"""
    v = records.ContributionDate(testlines)
    assert v.date.year == 2017
    assert v.date.month == 3
    assert v.date.day == 31


def test_NcdcRecord_chron_data(testchrondatalines):
    v = records.NcdcRecord(testchrondatalines)
    goal_datadict = {'depth': [1, 4], 'age': [2, 5], 'bacon': [3, 6]}
    goal_data = pd.DataFrame(goal_datadict)
    df_data = v.data.df.copy()
    for k in goal_datadict.keys():
        pd.testing.assert_series_equal(goal_data[k], df_data[k])

    goal_chrondict = {'Labcode': [152757], 'depth_top': [5], 'depth_bottom': [6],
                 'mat_dated': ['G. ruber or mixed planktonic'],
                 '14C_date': [475],'14C_1s_err': [30],
                 'delta_R': [188], 'delta_R_1s_err': [73],
                 'other_date': [pd.np.nan], 'other_1s_err': [pd.np.nan],
                 'other_type': [pd.np.nan]}
    goal_chron = pd.DataFrame(goal_chrondict)
    df_chron = v.chronology_information.df.copy()
    for k in goal_chrondict.keys():
        pd.testing.assert_series_equal(goal_chron[k], df_chron[k])



import logging

import numpy as np
import pandas as pd
import scipy.stats as stats
import snakebacon as sb
import carbonferret as cf


log = logging.getLogger(__name__)


def remove_outliers(df, col_name, iqr_min=10):
    """Remove outliers from dataframe copy based on col_name"""
    out = df.copy()
    iqr = stats.iqr(df.loc[:, col_name])
    if iqr < iqr_min:
        return out
    upper = np.percentile(df.loc[:, col_name], 75)
    lower = np.percentile(df.loc[:, col_name], 25)
    out = out[out.loc[:, col_name] < (upper + 1.5 * iqr)]
    out = out[out.loc[:, col_name] > (lower - 1.5 * iqr)]
    return out


def get_deltar_online(latlon, max_distance=3000):
    """Use carbonferret to grab an estimate ΔR from internet"""
    max_distance = int(max_distance)
    nearby = cf.find_near(lat=latlon[0], lon=latlon[1], n=10)

    nearby = nearby[nearby['distance (km)'] <= max_distance]
    # nearby = remove_outliers(nearby, 'DeltaR')
    # nearby = remove_outliers(nearby, 'DeltaRErr')
    log.debug('ΔR and ΔRσ from {} samples'.format(len(nearby)))

    w = 1/len(nearby)
    deltar_mean = nearby['DeltaR'].mean()
    # Use pooled or combined variance of carbonferret deltaR distributions.
    var = (w * ((nearby['DeltaR'] - deltar_mean)**2 + nearby['DeltaRErr']**2)).sum()
    sigma = np.sqrt(var)
    return tuple([deltar_mean, sigma])


def fit_agedepthmodel(chron, pdata, deltar=None, deltar_error=None, minyr=None, mcmc_kws=None):
    log.debug('Fitting new age model.')
    if minyr is None:
        minyr = -1000

    chron = pd.DataFrame(chron).copy()
    pdata = pd.DataFrame(pdata).copy()

    chron['depth'] = chron.loc[:, ('depth_top', 'depth_bottom')].mean(axis=1)
    chron['labid'] = chron.Labcode
    chron['age'] = chron.loc[:, '14C_date']
    chron['error'] = chron.loc[:, '14C_1s_err']

    chron['cc'] = 2  # This is marine calibration curve, use for all depths.

    chron.sort_values('depth', inplace=True)

    # Cleanup dataframe so works with snakebacon.
    if deltar is not None:
        chron['delta_R'] = deltar
    if deltar_error is not None:
        chron['delta_R_1s_err'] = deltar_error

    if chron.other_date.notnull().any():
        # Check that can't have 14C_date and other_date at same depth. Same for 1s_error.
        assert ~chron.loc[:, ['14C_date', 'other_date']].notnull().all(axis=1).any()
        assert ~chron.loc[:, ['14C_1s_err', 'other_1s_err']].notnull().all(axis=1).any()
        other_msk = ~chron['14C_date'].notnull()
        # `other_dates` should not have delta_R values
        chron.loc[other_msk, 'delta_R'] = 0
        chron.loc[other_msk, 'delta_R_1s_err'] = 0
        # Move `other_dates` and `errors` to chron.age and chron.error.
        chron.loc[other_msk, 'age'] = chron.loc[other_msk, 'other_date']
        chron.loc[other_msk, 'error'] = chron.loc[other_msk, 'other_1s_err']
        chron.loc[other_msk, 'cc'] = 0  # Using constant calibration curve for non-14C dates
        # Drop rows with `other_date` but no `other_1s_error`.
        chron.drop(chron[other_msk & chron['other_date'].notnull() & ~chron['other_1s_err'].notnull()].index, inplace=True)

    # Have any NaNs in age, depth or error?
    assert chron.loc[:, ['age', 'depth', 'error']].notnull().all().all()

    coredates = sb.ChronRecord(chron)

    d_min = np.min([x.min() for x in [chron.depth, pdata.depth]])
    d_max = np.max([x.max() for x in [chron.depth, pdata.depth]])
    sug_acc_mean = sb.suggest_accumulation_rate(coredates)
    # n_segs = np.ceil((d_max - d_min) / 5)  # Num segments in mcmc, ~ 5cm, rounded up.

    guesses = np.random.randn(2) * coredates.error[:2] + coredates.age[:2]  # TODO(brews): Check whether need sqrt error here.
    guesses[guesses < minyr] = minyr  # Line #70 of Bacon.R warns that otherwise twalk MCMC will not run.

    # if n_segs > 200 or n_segs < 5:
    # n_segs = np.ceil((d_max - d_min) / 10)
    # assert (n_segs < 500) and (n_segs > 5)  # Sanity check for extremely long or short MCMC runs.

    mcmc_params = dict(depth_min = d_min, depth_max = d_max,
                       cc = chron.cc.values,
                       d_r = chron.delta_R.values,
                       d_std = chron.delta_R_1s_err.values,
                       t_a =[3], t_b=[4],
                       k = 50,#n_segs,
                       minyr = minyr, maxyr = 50000,
                       th01 = guesses[0], th02 = guesses[1],
                       acc_mean = sug_acc_mean, acc_shape = 1.5,
                       mem_strength = 4, mem_mean = 0.7)

    if mcmc_kws is not None:
        mcmc_params.update(mcmc_kws)

    log.debug('MCMC parameters: {}'.format(mcmc_params))

    agemodel = sb.AgeDepthModel(coredates, mcmc_kws=mcmc_params)
    log.debug('Age model fit done.')
    return agemodel, coredates


def date_proxy(admodel, pdata, nsims):
    pdata = pd.DataFrame(pdata).copy()
    if 'age' in pdata.columns:
        # Rename to avoid error with sb.ProxyRecord
        log.debug('Renaming age column in proxy data')
        cols = list(pdata.columns.values)
        cols[cols.index('age')] = 'original_age'
        pdata.columns = cols

    orig_pdata = sb.ProxyRecord(pdata)
    pdata_median = admodel.date(orig_pdata, 'median').to_pandas()
    pdata_ensemble = admodel.date(orig_pdata, 'ensemble', nsims).to_pandas()
    pdata_median['age'] = pdata_median['age'].round()
    pdata_ensemble['age'] = pdata_ensemble['age'].round()
    return pdata_median, pdata_ensemble



import datetime
import copy
import logging

import numpy as np
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pylab as plt
import cartopy.feature as cfeature
import cartopy.crs as ccrs


log = logging.getLogger(__name__)


def write_reportpdf(qcplot_path, fldata, proxy_vars, latlon, deltar=None, deltar_error=None, proxys_median=None, agemodel=None):
    log.debug('Writing QC report plots')

    n_vars = len(proxy_vars)
    n_cols = n_vars + 1

    if deltar is None and agemodel is not None:
        deltar = agemodel.mcmcsetup.mcmc_kws['d_r'][0]
        deltar_error = agemodel.mcmcsetup.mcmc_kws['d_std'][0]

    with PdfPages(qcplot_path) as pdf:

        fig = plt.figure(figsize=(6.5, 9))

        ax2 = plt.subplot2grid((n_cols, 2), (0, 1),
                               projection=ccrs.Robinson(central_longitude=latlon[1]))
        ax3 = plt.subplot2grid((n_cols, 2), (0, 0))

        ax2 = site_map(latlon, ax=ax2)

        ax3 = summary_text(fldata, deltar_used=deltar, deltar_error_used=deltar_error, ax=ax3)

        for i, var in enumerate(proxy_vars):
            this_ax = plt.subplot2grid((n_cols, 1), (1 + i, 0), colspan=2)
            this_ax = proxyvar_timeseries(proxys_median, var, ax=this_ax)
            this_ax.xaxis.label.set_visible(False)
            this_ax.title.set_visible(False)
            if i == 0:
                this_ax.title.set_visible(True)
                this_ax.set_title('Proxy variables')
        this_ax.xaxis.label.set_visible(True)
        this_ax.set_xlabel('Age (cal yr BP)')
        fig.tight_layout()
        pdf.savefig(bbox_inches='tight')
        plt.close()

        fig = plt.figure(figsize=(6.5, 9))
        ax1 = plt.subplot2grid((3, 2), (0, 0))
        ax2 = plt.subplot2grid((3, 2), (0, 1))
        ax4 = plt.subplot2grid((3, 2), (1, 0), rowspan=2, colspan=2)

        if agemodel is not None:
            ax1 = sedrate(agemodel, ax=ax1)
            ax2 = sedmemory(agemodel, ax=ax2)
            ax4 = agedepth(agemodel, proxys_median, ax=ax4)

        fig.tight_layout()
        pdf.savefig(bbox_inches='tight')
        plt.close()

        fig = plt.figure(figsize=(6.5, 9))
        ax1 = plt.subplot(1, 1, 1)
        if agemodel is not None:
            ax1 = agedepth(agemodel, proxys_median, maxage=25000, ax=ax1)
        fig.tight_layout()
        pdf.savefig(bbox_inches='tight')
        plt.close()

        log.debug('QC report plot saved to {}'.format(qcplot_path))


def proxyvar_timeseries(proxys_median, var, ax=None):
    if ax is None:
        ax = plt.gca()
    # This is handled poorly on my part.
    if 'original_age' in proxys_median.columns:
        # ie...if has new agemodel data
        old_age = proxys_median.loc[:, ('original_age', var)].dropna()
        new_age = proxys_median.loc[:, ('age', var)].dropna()
        ax.plot(new_age.age, new_age[var], '.', color='C3', label='MCMC median age')
        ax.plot(new_age.age, new_age[var], color='C3', linewidth=0.5, label='_nolegend_')
        ax.plot(old_age.original_age, old_age[var], 'x', color='C0', label='File age')
        ax.plot(old_age.original_age, old_age[var], color='C0', linewidth=0.5, label='_nolegend_')
        log.debug('Found new agemodel proxy timeseries')
    else:
        old_age = proxys_median.loc[:, ('age', var)].dropna()
        ax.plot(old_age.age, old_age[var], 'x', color='C0', label='File age')
        ax.plot(old_age.age, old_age[var], color='C0', linewidth=0.5, label='_nolegend_')
        log.debug('Assuming no new agemodel proxy timeseries')

    if 'd18o' in var or 'd18O' in var:
        ax.invert_yaxis()
    ax.set_ylabel(var)
    ax.grid()
    return ax


def agedepth(agemodel, proxy_median, prior_dwidth=30, maxage=None, ax=None):
    """Plot comparing fit agemodel with age in the original proxy data"""
    if ax is None:
        ax = plt.gca()

    # This hack is going to break at some point.
    if maxage is None:
        maxage = agemodel.mcmcsetup.mcmc_kws['maxyr']
    if np.max(agemodel.age_median() > maxage):
        agemodel = copy.deepcopy(agemodel)
        too_old = ~(agemodel.age_median() > maxage)
        agemodel._depth = agemodel.depth[too_old]
        agemodel._age_ensemble = agemodel.age_ensemble[too_old]

    ax = agemodel.plot(ax=ax)
    ax.images[0].set_cmap(plt.cm.Greys)
    for l in ax.lines:
        l.set_color('C3')
    ax.plot(proxy_median.depth, proxy_median.original_age, 'C0', label='File age model')
    ax = agemodel.plot_prior_dates(dwidth=prior_dwidth, ax=ax)
    ax.collections[0].set_color('k')
    ax.collections[0].set_zorder(10)
    ax.autoscale_view()
    ax.set_title('Age model')
    return ax


def site_map(latlon, ax=None):
    """Simple global site map, given latlon"""
    if ax is None:
        ax = plt.gca(projection=ccrs.Robinson(central_longitude=latlon[1]))
    ax.set_global()
    ax.add_feature(cfeature.LAND, facecolor='#B0B0B0')
    ax.outline_patch.set_linewidth(0.5)
    ax.plot(latlon[1], latlon[0], 'o', color='C0', transform=ccrs.Geodetic())
    return ax


def summary_text(fldata, deltar_used=None, deltar_error_used=None, ax=None):
    if ax is None:
        ax = plt.gca()

    latlon = [fldata.site_information.northernmost_latitude,
              fldata.site_information.westernmost_longitude]

    elevation = fldata.site_information.elevation

    # These checks are not well written.
    try:
        deltar_original = np.int(np.round(fldata.chronology_information.df['delta_R'].values[0]))
        if np.isnan(deltar_original):
            deltar_original = None
    except (AttributeError, ValueError, IndexError) as e:
        deltar_original = None

    try:
        deltar_std_original = np.int(np.round(fldata.chronology_information.df['delta_R_1s_err'].values[0]))
        if np.isnan(deltar_std_original):
            deltar_std_original = None
    except (AttributeError, ValueError, IndexError) as e:
        deltar_std_original = None

    try:
        deltar_used = np.int(np.round(deltar_used))
    except (AttributeError, TypeError):
        deltar_used = None

    try:
        deltar_error_used = np.int(np.round(deltar_error_used, 2))
    except (AttributeError, TypeError):
        deltar_error_used = None


    text_template = 'Latitude: {}°\nLongitude: {}°\nElevation: {} m ' \
                    '\n\nΔR: {}\nΔRσ: {}\nFile ΔR: {}\nFile ΔRσ: {}'
    text_str = text_template.format(latlon[0], latlon[1], elevation,
                                    deltar_used, deltar_error_used,
                                    deltar_original, deltar_std_original)
    ax.text(0.05, 0.9, text_str, verticalalignment='top',
            horizontalalignment='left', transform=ax.transAxes)
    ax.set_title(fldata.site_name + '\n' + str(datetime.datetime.now().isoformat()))
    ax.axis('off')
    return ax


def sedrate(agemodel, ax=None):
    """Plot prior and posterior sediment rates for agemodel"""
    if ax is None:
        ax = plt.gca()
    ax = agemodel.plot_sediment_rate(ax)
    ax.lines[-1].set_color('C3')
    return ax


def sedmemory(agemodel, ax=None):
    """Plot prior and posterior sediment memory for agemodel"""
    if ax is None:
        ax = plt.gca()
    ax = agemodel.plot_sediment_memory(ax)
    ax.lines[-1].set_color('C3')
    return ax



import logging
import datetime
from io import BytesIO

import pandas as pd
from chardet import detect as chdetect


DIVIDER = '#------------------------'
DATALINE_TRIGGER = '# Data lines follow (have no #)'
HEADINGS = ['Contribution_Date', 'Title', 'Investigators', 
            'Description and Notes', 'Publication', 'Funding_Agency', 
            'Site Information', 'Data_Collection', 'Species', 
            'Chronology_Information', 'Variables', 'Data']
CHRON_HEADER = '# Labcode\tdepth_top\tdepth_bottom\tmat_dated\t14C_date\t14C_1s_err\tdelta_R\tdelta_R_1s_err\tother_date\tother_1s_err\tother_type\t'
MISSINGVALUE_LABEL = '# Missing Value: '


log = logging.getLogger(__name__)


def grab_latlon(g):
    """Grab latitude from proxychimp.Guts"""
    victim = g.pull_section('Site Information')
    fields = dict(Northernmost_Latitude=None,
                  Southernmost_Latitude=None,
                  Easternmost_Longitude=None,
                  Westernmost_Longitude=None)
    for ln in victim:
        if any([x in ln for x in fields.keys()]):
            ln_split = ln.split(':')
            v = float(ln_split[-1])
            k = ln_split[0][1:].lstrip().rstrip()
            assert k in fields.keys()
            fields[k] = v
    return fields['Northernmost_Latitude'], fields['Westernmost_Longitude']

def grab_elevation(g):
    """Grab site elevation from proxychimp.Guts"""
    victim = g.pull_section('Site Information')
    elevation = None
    for ln in victim:
        if 'Elevation:' in ln:
            ln_split = ln.split(':')
            elevation = float(ln_split[-1])
    return elevation

def grab_collection_year(g):
    """Get collection year from Data_Collection section of proxychimp.Guts"""
    victim = g.pull_section('Data_Collection')
    yr = None
    for ln in victim:
        if 'Collection_Year' in ln:
            ln_split = ln.split(':')
            yr = int(ln_split[-1])
    return yr

def grab_publication_year(g):
    """Get publication year from proxychimp.Guts"""
    victim = g.pull_section('Publication')
    yr = None
    for ln in victim:
        if 'Published_Date_or_Year' in ln:
            ln_split = ln.split(':')
            yr = int(ln_split[-1])
    return yr

def grab_contribution_date(g):
    """Get contribution date from proxychimp.Guts"""
    victim = g.pull_section('Contribution_Date')
    d = None
    for ln in victim:
        if 'Date' in ln:
            ln_split = ln.split(':')
            d_str = ln_split[-1].split('-')
            assert len(d_str) == 3
            d = datetime.date(int(d_str[0]), int(d_str[1]), int(d_str[2]))
    return d

def plot_sitemap(g):
    """Plot site location on global map from proxychimp.Guts"""
    pass


def plot_agedepth(g):
    """Plot given age-depth relationship from proxychimp.Guts data"""
    pass



class Guts:
    """Ghetto and error-tolerant way to parse sections of NCDC proxy text files
    """

    def __init__(self, str_or_path):
        self._section_headings = ['# ' + s for s in HEADINGS]

        self.path = None
        self.encodingguess = None
        self.filestr = None
        try:
            with open(str_or_path, 'rb') as fl:
                flbytes = fl.read()
            self.path = str(str_or_path)
            self.encodingguess = chdetect(flbytes)
            # TODO(brews): Not sure we need ~both~ filestr, and lines + self.data + self.description.
            self.filestr = flbytes.decode(self.encodingguess['encoding'])
        except (FileNotFoundError, OSError):
            self.filestr = str(str_or_path)
        lines = self.filestr.splitlines()

        self.data = []
        self.description = []
        self.data_beginline = None
        self._divide_portions(lines)

        self.sectionindex = None
        self._write_sectionindex()

    def _divide_portions(self, lines):
        """Divide guts into 'description' and 'data' sections
        """
        prev_description = True
        dataline_flag = False
        for n, ln in enumerate(lines):
            if ln[0] == '#' or dataline_flag is False:
                self.description.append(ln)
                if DATALINE_TRIGGER in ln:
                    dataline_flag = True
                    log.debug('Found dataline flag on line {0}'.format(n))
            elif dataline_flag:
                self.data.append(ln)
                if prev_description is True:
                    self.data_beginline = n
                    log.debug('Data portion begins on line {0}'.format(n))
                prev_description = False

    def _write_sectionindex(self):
        """Populate the section index"""
        all_keys = []
        all_start = []
        prev_divider = False
        for idx, ln in enumerate(self.description):  # Lines after 6-line title.
            if prev_divider is True and any([h == ln.rstrip() for h in self._section_headings]):
                key = ln[1:].rstrip().lstrip()
                all_keys.append(key)
                all_start.append(idx)
                prev_divider = False
            if DIVIDER in ln:
                prev_divider = True
        all_stop = [x - 1 for x in all_start[1:]]
        all_stop.append(self.data_beginline)
        self.sectionindex = {k: (v0, v1) for (k, v0, v1) in zip(all_keys, all_start, all_stop)}

    def pull_section(self, section):
        """Grab a list of line strings from the file description"""
        return self.description[slice(*self.sectionindex[section])]

    def available_sections(self):
        """Get a list of available sections in the file"""
        return list(self.sectionindex.keys())

    def yank_data_df(self):
        """Get 'data' information as dataframe"""
        lines = [x.rstrip() for x in self.data]
        data_bytes = '\n'.join(lines).encode('utf-8')
        missingvalues = self.guess_missingvalues()
        df = pd.read_table(BytesIO(data_bytes), na_values=missingvalues)
        # df = pd.read_table(BytesIO(data_bytes), na_values=[-999, 'NaN'])
        return df

    def yank_chron_df(self, section_name='Chronology_Information', missingvalues=None):
        """Get chronology information as dataframe"""
        if missingvalues is None:
            missingvalues = [-999, 'NaN']
        section = self.pull_section(section_name)
        start_idx = section.index(CHRON_HEADER)
        g_chrond = section[start_idx:]
        g_chrond_cleaned = [x[2:].rstrip() for x in g_chrond]  # Removes the '# ' and ending white space.
        data_bytes = '\n'.join(g_chrond_cleaned).encode('utf-8')
        df = pd.read_table(BytesIO(data_bytes), na_values=missingvalues)
        return df

    def guess_missingvalues(self):
        """Guess data section missing values"""
        section = self.pull_section('Data')
        out = None
        for ln in section:
            if MISSINGVALUE_LABEL in ln:
                out = ln.split(MISSINGVALUE_LABEL)[1]
        log.debug('Guessed missing value(s): {}'.format(out))
        return out

    def has_data(self):
        """Check if has populated data information"""
        try:
            d = self.yank_data_df()
        except KeyError:
            return False
        if len(d) > 0:
            result = True
        else:
            result = False
        return result

    def has_chron(self):
        """Check if has populated chronology information"""
        try:
            chron = self.yank_chron_df()
        except KeyError:
            return False
        if len(chron) > 0:
            result = True
        else:
            result = False
        return result

    def has_deltar(self):
        """Check if has populated delta R chronology information"""
        try:
            chron = self.yank_chron_df()
        except KeyError:
            return False
        if any(chron.delta_R.notnull()):
            result = True
        else:
            result = False
        return result

    def has_deltar_error(self):
        """Check if has populated delta R error chronology information"""
        try:
            chron = self.yank_chron_df()
        except KeyError:
            return False
        if any(chron.delta_R_1s_err.notnull()):
            result = True
        else:
            result = False
        return result

    def has_datacolumn(self, name):
        """Check if name is in data section columns"""
        try:
            data = self.yank_data_df()
        except KeyError:
            return False
        if name in data.columns:
            result = True
        else:
            result = False
        return result



from setuptools import setup, find_packages


setup(
    name='proxysiphon',
    version='0.0.1a0',
    description='Internal lab tool to parse and clean marine sediment proxy data.',
    license='GPLv3',

    author='S. Brewster Malevich',
    author_email='malevich@email.arizona.edu',
    url='https://github.com/brews/proxysiphon',
    classifiers=[
        'Development Status :: 1 - Planning',

        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',

        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        'Programming Language :: Python :: 3',
    ],
    keywords='marine paleoclimate',

    packages=find_packages(exclude=['docs']),

    install_requires=['numpy', 'scipy', 'matplotlib', 'pandas', 'cartopy', 'attrs',
                      'carbonferret', 'snakebacon', 'erebusfall'],
    tests_require=['pytest'],
    package_data={'proxysiphon': ['tests/*.txt']},
)
